use crate::vm::source_code_location::SourceCodeLocation;
use pest::Parser;

pub mod build;

#[derive(Debug)]
pub struct CompilationUnit {
    namespace: Option<Fqn>,
    declarations: Vec<Declaration>,
}

impl CompilationUnit {
    pub fn namespace(&self) -> &Option<Fqn> {
        &self.namespace
    }

    pub fn declarations(&self) -> &Vec<Declaration> {
        &self.declarations
    }
}

#[derive(Debug)]
pub struct Fqn {
    identifiers: Vec<Identifier>,
}

impl Fqn {
    pub fn new(identifiers: Vec<Identifier>) -> Self {
        Fqn { identifiers }
    }

    pub fn identifiers(&self) -> &Vec<Identifier> {
        &self.identifiers
    }

    pub fn push_identifier(&mut self, identifier: Identifier) {
        self.identifiers.push(identifier);
    }

    pub fn pop_identifier(&mut self) {
        self.identifiers.pop();
    }
}

#[derive(Debug)]
pub enum Declaration {
    Interface {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        type_declaration: TypeDeclaration,
    },
    Implementation {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        type_declaration: TypeDeclaration,
        impl_ctor: Option<ClassConstructor>,
    },
    Module {
        identifier: Identifier,
        is_extern: bool,
        is_public: bool,
        declarations: Vec<Declaration>,
    },
    Function(FunctionDeclaration),
    Prop(Prop),
    Field(Field),
}

#[derive(Debug)]
pub struct FunctionDeclaration {
    source_code_location: SourceCodeLocation,
    identifier: Identifier,
    is_extern: bool,
    is_public: bool,
    parameters: Vec<Parameter>,
    declared_type: Option<TypeUse>,
    block_statement: BlockStatement,
}

impl FunctionDeclaration {
    pub fn source_code_location(&self) -> &SourceCodeLocation {
        &self.source_code_location
    }

    pub fn identifier(&self) -> &Identifier {
        &self.identifier
    }

    pub fn block_statement(&self) -> &BlockStatement {
        &self.block_statement
    }
}

#[derive(Debug)]
pub struct TypeDeclaration {
    generics: Vec<GenericParameter>,
    extends: Vec<TypeUse>,
    declarations: Vec<Declaration>,
}

impl TypeDeclaration {
    pub fn new(
        generics: Vec<GenericParameter>,
        extends: Vec<TypeUse>,
        declarations: Vec<Declaration>,
    ) -> TypeDeclaration {
        TypeDeclaration {
            generics,
            extends,
            declarations,
        }
    }
}

#[derive(Debug, Clone)]
pub struct Identifier {
    name: String,
}

impl Identifier {
    pub fn name(&self) -> &str {
        &self.name
    }
}

#[derive(Debug)]
pub struct GenericParameter {
    identifier: Identifier,
    bound: Option<GenericBound>,
}

#[derive(Debug)]
pub enum GenericBound {
    Extends { fqn: Fqn },
    Super { fqn: Fqn },
}

#[derive(Debug)]
pub struct GenericArgument {
    fqn: Fqn,
}

#[derive(Debug)]
pub struct TypeUse {
    fqn: Fqn,
    generics: Vec<GenericArgument>,
}

#[derive(Debug)]
pub struct ClassConstructor {
    args: Vec<ClassMember>,
}

impl ClassConstructor {
    pub fn new(args: Vec<ClassMember>) -> Self {
        ClassConstructor { args }
    }
}

#[derive(Debug)]
pub struct ClassMember {
    is_field: bool,
    identifier: Identifier,
    type_use: Option<TypeUse>
}

impl ClassMember {
    pub fn new(is_field: bool, identifier: Identifier, type_use: Option<TypeUse>) -> Self {
        ClassMember {
            is_field,
            identifier,
            type_use
        }
    }
}

#[derive(Debug)]
pub struct ImplCtorArg {
    is_field: bool,
    identifier: Identifier,
    r#type: Option<TypeUse>,
}

#[derive(Debug)]
pub struct Prop {
    identifier: Identifier,
    r#type: TypeUse,
}

#[derive(Debug)]
pub struct Field {
    identifier: Identifier,
    r#type: TypeUse,
}

#[derive(Debug)]
pub struct BlockStatement {
    statements: Vec<Statement>,
}

impl BlockStatement {
    pub fn statements(&self) -> &Vec<Statement> {
        &self.statements
    }
}

#[derive(Debug)]
pub enum Statement {
    CallStatement(CallExpression),
    AssignStatement(AssignExpression),
}

#[derive(Debug)]
pub struct CallExpression {
    receiver: Expression,
    actuals_list: Vec<Expression>,
    source_code_location: SourceCodeLocation,
}

impl CallExpression {
    pub fn receiver(&self) -> &Expression {
        &self.receiver
    }

    pub fn actuals_list(&self) -> &Vec<Expression> {
        &self.actuals_list
    }

    pub fn source_code_location(&self) -> &SourceCodeLocation {
        &self.source_code_location
    }
}

#[derive(Debug)]
pub struct AssignExpression {
    left: LValue,
    right: Expression,
}

#[derive(Debug)]
pub enum Expression {
    CallExpression(Box<CallExpression>),
    AssignExpression(Box<AssignExpression>),
    LValue(Box<LValue>),
    Literal(Box<Literal>),
}

#[derive(Debug)]
pub enum LValue {
    Fqn(Fqn),
}

#[derive(Debug)]
pub enum Literal {
    IntegerLiteral(i32),
    LongLiteral(i64),
    DoubleLiteral(f64),
    USizeLiteral(usize),
    StringLiteral(String),
    BooleanLiteral(bool),
}

#[derive(Debug)]
pub struct Parameter {
    identifier: Identifier,
    declared_type: Option<TypeUse>,
}
